var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import * as React from "react";
import { FLOW_UPDATE_EVENT, getKey, makeKey, } from "../lib";
const flowContext = React.createContext(null);
export function useController() {
    return React.useContext(flowContext).controller;
}
export function usePrefetcher() {
    const controller = useController();
    const keys = React.useRef(new Set()).current;
    const rerender = React.useState()[1];
    React.useEffect(() => {
        return controller.on(FLOW_UPDATE_EVENT, (flow) => {
            if (!keys.has(flow.key))
                return;
            rerender({});
        });
    }, [keys, rerender, controller]);
    return React.useMemo(() => [
        (...args) => {
            const key = Array.isArray(args[0]) ? makeKey(args[0]) : args[0];
            keys.add(key);
            const flow = controller.start(...args);
            return flow;
        },
        (key, ...args) => {
            const [fn, mapper, defaultValue] = typeof args[0] === "function" && typeof args[1] === "function"
                ? args
                : [undefined, ...args];
            const k = Array.isArray(key) ? makeKey(key) : key;
            const flow = fn ? controller.flow(k, fn) : controller.flow(k);
            return flow ? mapper(flow) : defaultValue;
        },
    ], [controller, keys]);
}
export function useFlow(...args) {
    // useFlow(options)
    if (args[0] &&
        args.length === 1 &&
        typeof args[0] === "object" &&
        !Array.isArray(args[0])) {
        const _a = args[0], { flow, key } = _a, options = __rest(_a, ["flow", "key"]);
        args = [key, flow, options];
    }
    let fixedArgs = false;
    const originalKey = args[0];
    // key array  useFlow([key, arg1, arg2, arg3, arg4])
    if (Array.isArray(args[0])) {
        // overwrite args option
        // exclude first item (key)
        args[2] = Object.assign(Object.assign({}, args[2]), { args: args[0].slice(1) });
        args[0] = makeKey(args[0]);
        fixedArgs = true;
    }
    const provider = React.useContext(flowContext);
    if (!provider) {
        throw new Error("No flow provider found");
    }
    const prependArgs = [];
    let overrideArgs;
    if (typeof args[1] === "function") {
        overrideArgs = args;
    }
    // useFlow(fn, options)
    else if (typeof args[0] === "function") {
        overrideArgs = [getKey(args[0]), args[0], args[1]];
    }
    else {
        // useFlow(key, options)
        if (Array.isArray(originalKey)) {
            prependArgs.push(...originalKey);
        }
        else {
            prependArgs.push(originalKey);
        }
        let defaultFlow;
        const defaultFlowKey = prependArgs[0];
        if (typeof provider.defaultFlow === "function") {
            defaultFlow = provider.defaultFlow;
        }
        else if (typeof provider.defaultFlow === "object" &&
            defaultFlowKey in provider.defaultFlow) {
            defaultFlow = provider.defaultFlow[defaultFlowKey];
            prependArgs.shift();
        }
        else {
            throw new Error(`No default flow provided for key ${originalKey}`);
        }
        overrideArgs = [getKey(args[0]), defaultFlow, args[1]];
    }
    const [key, fn, _b] = overrideArgs, _c = _b === void 0 ? {} : _b, { suspense = provider.suspense, errorBoundary = provider.errorBoundary } = _c, options = __rest(_c, ["suspense", "errorBoundary"]);
    const rerender = React.useState()[1];
    const optionsRef = React.useRef({});
    const renderingRef = React.useRef(false);
    const flowRef = React.useRef();
    const { flowHook, handleSuspeseAndErrorBoundary } = React.useMemo(() => {
        return createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs);
    }, [suspense, errorBoundary, fixedArgs]);
    // update refs
    renderingRef.current = true;
    optionsRef.current = Object.assign(Object.assign({}, options), { suspense,
        errorBoundary,
        prependArgs });
    // make sure removed flow does not cause an error
    flowRef.current =
        provider.controller.flow(key, fn) || flowRef.current;
    React.useLayoutEffect(() => {
        var _a;
        renderingRef.current = false;
        const status = (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.status;
        return provider.controller.on(FLOW_UPDATE_EVENT, (flow) => {
            var _a, _b, _c, _d, _e, _f;
            if (optionsRef.current.disabled)
                return;
            if (flow.key === key) {
                rerender({});
                if (flow.status !== status) {
                    (_b = (_a = optionsRef.current).onUpdated) === null || _b === void 0 ? void 0 : _b.call(_a, flow);
                    if (flow.completed) {
                        (_d = (_c = optionsRef.current).onCompleted) === null || _d === void 0 ? void 0 : _d.call(_c, flow.data);
                    }
                    else if (flow.faulted) {
                        (_f = (_e = optionsRef.current).onFaulted) === null || _f === void 0 ? void 0 : _f.call(_e, flow.error);
                    }
                }
            }
        });
    });
    handleSuspeseAndErrorBoundary();
    return flowHook;
}
function createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs) {
    function handleSuspeseAndErrorBoundary() {
        if (!flowRef.current || !renderingRef.current)
            return;
        if (optionsRef.current.errorBoundary &&
            flowRef.current.status === "faulted") {
            throw flowRef.current.error;
        }
        if (optionsRef.current.suspense && flowRef.current.status === "running") {
            throw new Promise((resolve) => {
                var _a;
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.on("update", resolve);
            });
        }
    }
    function getArgs(inputArgs) {
        const prependArgs = optionsRef.current.prependArgs || [];
        const defaultArgs = optionsRef.current.args;
        if (!defaultArgs) {
            return [...prependArgs, ...inputArgs];
        }
        return [
            ...prependArgs,
            ...inputArgs,
            ...defaultArgs.slice(inputArgs.length),
        ];
    }
    function run(type, args) {
        var _a;
        if (optionsRef.current.disabled) {
            return flowHook;
        }
        if (args.length && fixedArgs) {
            throw new Error("Passing arguments to fixed args flow is not allowed. Use the overload useFlow(key, flow, { args: [] }) instead");
        }
        (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a[type](...getArgs(args));
        handleSuspeseAndErrorBoundary();
        return flowHook;
    }
    const flowHook = {
        get hasData() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.hasData) || false;
        },
        get status() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.status) || "unknown";
        },
        get error() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.error;
        },
        get running() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.running) || false;
        },
        get completed() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.completed) || false;
        },
        get faulted() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.faulted) || false;
        },
        get idle() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.idle) || false;
        },
        get data() {
            var _a, _b;
            if (!((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.hasData)) {
                return optionsRef.current.defaultData;
            }
            return (_b = flowRef.current) === null || _b === void 0 ? void 0 : _b.data;
        },
        get disabled() {
            return optionsRef.current.disabled || false;
        },
        get current() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.current;
        },
        get cancelled() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancelled) || false;
        },
        get stale() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.stale) || false;
        },
        update(data) {
            var _a;
            if (!optionsRef.current.disabled) {
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.update(data);
            }
            return flowHook;
        },
        get(...args) {
            return flowHook.start(...args).data;
        },
        tryGet(defaultValue, ...args) {
            const { result = defaultValue } = {
                result: flowHook.start(...args).data,
            };
            return result;
        },
        start(...args) {
            return run("start", args);
        },
        restart(...args) {
            return run("restart", args);
        },
        next(payload) {
            var _a;
            if (!optionsRef.current.disabled) {
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.next(payload);
            }
            return flowHook;
        },
        cancel() {
            var _a;
            if (!optionsRef.current.disabled) {
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancel();
            }
            return flowHook;
        },
    };
    return { flowHook, handleSuspeseAndErrorBoundary };
}
export const FlowProvider = (props) => {
    const parentProvider = React.useContext(flowContext);
    const rerender = React.useState()[1];
    const { controller = parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.controller, suspense = (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.suspense) || false, errorBoundary = (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.errorBoundary) || false, defaultFlow = parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.defaultFlow, fallback, children, } = props;
    const value = React.useMemo(() => ({ controller, suspense, errorBoundary, defaultFlow }), [controller, suspense, errorBoundary, defaultFlow]);
    if (!controller) {
        throw new Error("No controller provided");
    }
    if (controller.error) {
        throw controller.error;
    }
    if (!controller.ready) {
        if (suspense)
            throw controller.promise;
        controller.promise.finally(() => rerender({}));
        return React.createElement(React.Fragment, {}, fallback);
    }
    return React.createElement(flowContext.Provider, {
        value,
        children,
    });
};
